; keypad_encoder_606.a51  HC05 code translated to 8051
; Add ALE assertion for ROM run on big board, alter MATCH label after assy failed (11/13)
; 5/10/06, keytable revised to reverse-top-bottom or left-right stimulus pattern
; second revision: hold inputs fixed for four, scan responses, for 0..3, etc.
;  (after looking at how keypad is connected: stimulus drives horizontal (rows)
;  try placing low in memory after ROM version fails for left column

; changes: put out zeros rather than highs, as stimulus

	$NOSYMBOLS
	$INCLUDE (C:\MICRO\8051\RAISON\INC\REG320.INC)
	$INCLUDE (C:\MICRO\8051\RAISON\INC\VECTORS320.INC)
	
; PORT USE:
;	P1: high 4 = outputs, low 4 = inputs
;   P3.4 strobes display's EN* (useful so we don't see the scanning values)
	
	
	STACKBOT EQU 80h	; Not necessary, but a good place for it
	ANYKEY EQU 0Fh		; simplest output stimulus value, used to test whether ANY key is down
	SAVE_LO4 EQU 0Fh	; mask to keep response, clearing stimulus
	NOKEY EQU 0Fh		; input indicating no key down
	LASTPLUS1 EQU 47h		; last entry in stimulus table, plus one (used to test for finish of scan)

	OUTCNT EQU 80h                ; for approx 10ms delay, (sets number of outer loops
                                ; used in DELAY routine)
                                ; --while debugging, set to 1
                                ; increased from 10h to 80h for flash version, after first-column
                                ;  keys fail (all deliver zero), whereas it works with external bus
                                ;  (my theory is that the 4X faster execution from internal mem messes up execution)
                                ;  (slower crystal did not solve problem)

;;;;;NOW PROGRAM CODE:
	ORG 0H
	AJMP BEGIN

        ORG 180h         

BEGIN:   MOV 0C4h, #04h	; turn on ALE during ROM run (bit 2, ALEON, of PMR register, C4h)
         ACALL INIT		; go off and set things up
ANY:     MOV P1, #ANYKEY    ; set up 4 lines low, to keypad: here we ask
								; the crude question, "Is ANY key down?"

CHECK:   MOV A, P1       ; read 4-bit input value (low 4 bits)
        ANL A, #SAVE_LO4     ; ...keep the low 4 bits

        CJNE A,#0Fh, PROCEED_MB    ; if low 4 aren't all high, then some key has been pressed
        SJMP  CHECK       ; But we land here if nothing is pressed.  Go look again 


PROCEED_MB:      	ACALL DELAY     ; check for bounce:
      	MOV A, P1       ; read 4-bit input value again
         ANL A, #SAVE_LO4 	; If key still down, good
        CJNE A,#NOKEY, PROCEED_SURE    ; if low 4 still aren't all high, OK
        SJMP  CHECK       ; ...but if not still down, it's just bounce.  Look again.
      	
      	  ; Now a key is down.  Go scan for particular pattern                  
 
PROCEED_SURE: 		MOV R2, #1		;  start offset at first OUT entry; we'll walk through the table         

TRYKEY:  ACALL GETCODE    ; Get an OUT pattern from table; send 1 bit hi, 
        MOV P1, A         ; ...send stimulus: just one bit LOW; low 4 all high (to allow P1 to serve as input on those)
        MOV A, P1			;  pick up returned value (low 4)
        ANL A, #SAVE_LO4		;  ...knocking out pattern that was sent, keeping single input bit
        MOV R0, A			; save input bit pattern that was sensed
        INC R2				; Now point to possible responses (test for any single bit low, on low 4)
        ACALL GETCODE
								; Now see if value returned (low 4) matches table entry
        CJNE A, 0, NOFITS      ; ("0" is address of saved 4-bit response, saved in R0)
                                ;  predicted in table
                            
        SJMP FITS         ; If matches, go pick up key value from table		
        
NOFITS:    INC R2            ; ...but if no match, move down table
        INC R2            ;  (2 steps, because each entry includes a second
                        ;  byte: the key value)
        MOV A, R2
        CJNE A, #LASTPLUS1, TRYKEY   ; ...and go pick up a new trial value, if table not done
        						; (table is 48 long, but last entry, key value, we'd not use if no match)
                        
        SJMP ANY         ; If we land here, NO key pattern appeared: No
                        ;  key was down; must have been bounce, so go start
                        ;  the scanning process afresh

FITS:  INC R2			; advance index pointer by one, to pick up key value from table

		ACALL GETCODE  ; this routine, fed offset in Acc, returns        
                        ; key value from table, one beyond
                        ;  stimulus-&-response pair 

        ACALL SHOWIT      ; This sends key value on 4-bit "bus"--to keypad (harmless)
                        ;  and also to HP display
        MOV P1, #0Fh   ; Wait till key is released before proceeeding: process is to send 
        						;  original all-LOWs on top 4 bits, and watch for key-down:

STILLKEY: MOV A, P1       ; read the 4 input bits
        ANL A, #SAVE_LO4     ; ...keep the low 4 bits
        CJNE A,#NOKEY, STILLKEY    ; if low 4 aren't all high, then loop till release
        ACALL DELAY      ; allow time for bounce
        SJMP ANY         ; Now key is released, so go scan anew for next
                        ;  key pressing 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;INITIALIZATIONS

INIT:    SETB P3.4			; start with EN disasserted
			RET
			
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
			ORG 100h
DELAY:	PUSH ACC				; save registers that'll get messed up
			PUSH B

			MOV A,#OUTCNT		; get outer-loop delay value
INITINNER:	MOV B,#0			;initialize inner loop counter, for max day
INLOOP:	DJNZ B, INLOOP		; count down inner loop, till inner hits zero
			DJNZ ACC,INITINNER	; ...then dec outer, and start inner again.
			POP B  				; restore saved registers
			POP ACC
			RET					; Now back to main program.

;;;;;;;;;;;;;;;;;;;;;;;
; SHOWIT

SHOWIT:  SWAP A			; swap nybbles, to get display value into display position
			ORL A, #0Fh		; force low nybble high, not to disturb inputs
			MOV P1, A
        ACALL PULSE       ; uses a tiny routine to pulse display latch
                        ;  (slow, but tidy to write & use)
; next two lines just to aid debugging: show on 8000 display
;		MOV DPTR, #8000h
;		MOVX @DPTR, A                        
        RET

;;;;;;;;;

; PULSE
; pulses display's Enable; pin, active low

PULSE:  CLR P3.4    ; take EN; low
        SETB P3.4   ; ...and then high again
        RET

;;;;;;;;;;;;;;
; KEY TABLE: SHOWS ALL COMBIS OF OUTPUT AND READBACK, 4-BITS EACH,
;  AND FOR EACH COMBINATION PROVIDES A SINGLE DESIRED KEY "VALUE"
GETCODE: 	MOV A, R2		; recall saved index value, for use in indexed read

			MOVC A, @A+PC		;  that we must read beyond
			RET

; the following table assumes active LOW stimulus--so just one input bit will be low
KEYTBL:  DB     07Fh,07h,0          ; top left corner: key value is zero
                              ;  ("DB" assigns storage, as many as needed
                              ; always send low nybble hi, to allow these bits as inputs)
        DB     07Fh,0Bh,1           ; 2d column, 1st row
        DB     07Fh,0Dh,2           ; 3d column, 1st row
        DB     07Fh,0Eh,3           ; 4th column,1st row
        DB     0BFh,07h,4           ; 1 col, 2 row
        DB     0BFh,0Bh,5				 
        DB     0BFh,0Dh,6
        DB     0BFh,0Eh,7
        DB     0DFh,07h,8				; 1 col, 3 row
        DB     0DFh,0Bh,9
        DB     0DFh,0Dh,0Ah
        DB     0DFh,0Eh,0Bh
        DB     0EFh,07h,0Ch			; 1 col, 4 row
        DB     0EFh,0Bh,0Dh
        DB     0EFh,0Dh,0Eh
        DB     0EFh,0Eh,0Fh           ; 4th column, 4th row (bot right)			


	END
